// This file contains your Data Connector logic
section Delika;

base_url = Extension.LoadString("BaseUrl");
client_id = Extension.LoadString("OAuthClientId");
client_secret = Extension.LoadString("OAuthClientSecret");
scope = Extension.LoadString("OAuthScope");
policy = Extension.LoadString("OAuthPolicy");
redirect_uri = Extension.LoadString("OAuthRedirectUri");
windowHeight = 600;
windowWidth = 600;

[DataSource.Kind="Delika", Publish="Delika.Publish"]
shared delika.Data = Value.ReplaceType(delika.DataImpl, delika.DataType);

delika.DataType =  type function (
    account as (type text meta [
        Documentation.FieldCaption = "Account",
        Documentation.FieldDescription = "Text of account name",
        Documentation.SampleValues = {"account01"}
    ]),
    dataset as (type text meta [
        Documentation.FieldCaption = "Dataset",
        Documentation.FieldDescription = "Text of dataset name",
        Documentation.SampleValues = {"dataset01"}
    ]),
    data as (type text meta [
        Documentation.FieldCaption = "Data",
        Documentation.FieldDescription = "Text of data name",
        Documentation.SampleValues = {"sample.csv"}
    ])
    )
    as table meta [
        Documentation.Name = Extension.LoadString("DataSourceLabel"),
        Documentation.LongDescription = Extension.LoadString("DataSourceLabel")
    ];

delika.DataImpl = (account as text, dataset as text, data as text) =>
   let
      schemaResult = Json.Document(Web.Contents(base_url & "/data/" & account & "/" & dataset & "/" & data & "/metadata/schema" )),
      schema = Json.Document(schemaResult[Data][Schema]),
      source = Binary.Decompress(
              Web.Contents(base_url & "/data/" & account & "/" & dataset & "/" & data & "/rawdata" ),
              Compression.GZip
              ) ,
      csv =
         if
            schema[DataSpecification][DataOptions][Header] = true
         then
            let
               columns =
                  [
                     Delimiter = schema[DataSpecification][DataOptions][FieldDelimiter],
                     Encoding = GetTextEncoding(schema[DataSpecification][DataOptions][Encoding])
                  ],
               sourceCsv = Csv.Document(source,columns),
               promoted = Table.PromoteHeaders(sourceCsv),
               originalColumns = Table.ColumnNames(promoted),
               schemaColumns = List.Transform(schema[TableDefinition][Columns], each _[Name]),
               renamed =
                  Table.RenameColumns(
                     promoted,
                     List.Transform(
                        List.Positions(originalColumns),
                        each {
                           List.Single(List.Range(originalColumns,_,1)),
                           List.Single(List.Range(schemaColumns,_,1))
                        }
                     )
                  )
            in
               renamed
         else
            let
               columns =
                  [
                     Columns = List.Transform(schema[TableDefinition][Columns], each _[Name]),
                     Delimiter = schema[DataSpecification][DataOptions][FieldDelimiter],
                     Encoding = GetTextEncoding(schema[DataSpecification][DataOptions][Encoding])
                  ]
            in
               Csv.Document(source, columns),
      replaced =
         Table.ReplaceValue(
            csv,
            schema[DataSpecification][DataOptions][NullString],
            null,
            Replacer.ReplaceValue,
            Table.ColumnNames(csv)),
      columnTypes =
         List.Transform(
            schema[TableDefinition][Columns],
            each  { _[Name], GetPowerQueryType(_[DataType][DataType])}
         ),
      changed = Table.TransformColumnTypes(replaced,columnTypes)
   in
      changed;

[DataSource.Kind="Delika.SQL", Publish="Delika.SQL.Publish"]
shared delika.Sql = Value.ReplaceType(delika.SqlImpl, delika.SqlType);

delika.SqlType =  type function (
    sql as (type text meta [
        Documentation.FieldCaption = "Sql",
        Documentation.FieldDescription = "Text of SQL",
        Documentation.SampleValues = {"SELECT t.id, t.value1, t.value2 WHERE [account01/dataset01/sample.csv] t"},
        Formatting.IsMultiLine = true,
        Formatting.IsCode = true
    ]))
    as table meta [
        Documentation.Name = Extension.LoadString("DataSourceLabelSql"),
        Documentation.LongDescription = Extension.LoadString("DataSourceLabelSql")
    ];


shared delika.SqlImpl = (query as text) =>
   let
      runQueryResult = RunQuery(query),
      queryId = runQueryResult[Data][QueryId],
      getQueryStatusResult = GetQueryStatus(queryId),
      schemaResult = if getQueryStatusResult[Data][Status] = "Finished" then Json.Document(GetQuerySchema(queryId)) else Error.Unexpected(),
      schema = Json.Document(schemaResult[Data][Schema]),
      queryResult = GetQueryResults(queryId),
      columns =
         [
            Columns = List.Transform(schema[TableDefinition][Columns], each _[Name]),
            Delimiter = schema[DataSpecification][DataOptions][FieldDelimiter],
            Encoding = GetTextEncoding(schema[DataSpecification][DataOptions][Encoding])
         ],
      csv = Csv.Document(queryResult, columns),
      columnTypes =
         List.Transform(
            schema[TableDefinition][Columns],
            each  { _[Name], GetPowerQueryType(_[DataType][DataType])}
         ),
      changed = Table.TransformColumnTypes(csv,columnTypes)
   in
      changed;


[DataSource.Kind="Delika.QueryResults", Publish="Delika.QueryResults.Publish"]
shared delika.QueryResults = Value.ReplaceType(delika.QueryResultsImpl, delika.QueryResultsType);

delika.QueryResultsType =  type function (
    queryId as (type text meta [
        Documentation.FieldCaption = "Query ID",
        Documentation.FieldDescription = "Identifier of query result",
        Documentation.SampleValues = {"673cd50d-ae01-4971-8521-bb8755590c68"}
    ]))
    as table meta [
        Documentation.Name = Extension.LoadString("DataSourceLabelQueryResults"),
        Documentation.LongDescription = Extension.LoadString("DataSourceLabelQueryResults")
    ];

shared delika.QueryResultsImpl = (queryId as text) =>
   let
      getQueryStatusResult = GetQueryStatus(queryId),
      schemaResult = if getQueryStatusResult[Data][Status] = "Finished" then Json.Document(GetQuerySchema(queryId)) else Error.Unexpected(),
      schema = Json.Document(schemaResult[Data][Schema]),
      queryResult = GetQueryResults(queryId),
      columns =
         [
            Columns = List.Transform(schema[TableDefinition][Columns], each _[Name]),
            Delimiter = schema[DataSpecification][DataOptions][FieldDelimiter],
            Encoding = GetTextEncoding(schema[DataSpecification][DataOptions][Encoding])
         ],
      csv = Csv.Document(queryResult, columns),
      columnTypes =
         List.Transform(
            schema[TableDefinition][Columns],
            each  { _[Name], GetPowerQueryType(_[DataType][DataType])}
         ),
      changed = Table.TransformColumnTypes(csv,columnTypes)
   in
      changed;

GetTextEncoding = (e as text) =>
   let
      encoding = Text.Lower(e)
   in
      if encoding = "utf-8" then TextEncoding.Utf8
      else if encoding = "utf-16" then TextEncoding.Utf16
      else if encoding = "us-ascii" then 20127
      else if encoding = "utf-7" then 65000
      else if encoding = "iso-8859-1" then 28591
      else if encoding = "utf-32" then 12000
      else if encoding = "utf-32be" then 12001
      else if encoding = "utf-16be" then 1201
      else if encoding = "ibm037" then 37
      else if encoding = "ibm437" then 437
      else if encoding = "ibm500" then 500
      else if encoding = "asmo-708" then 708
      else if encoding = "dos-720" then 720
      else if encoding = "ibm737" then 737
      else if encoding = "ibm775" then 775
      else if encoding = "ibm850" then 850
      else if encoding = "ibm852" then 852
      else if encoding = "ibm855" then 855
      else if encoding = "ibm857" then 857
      else if encoding = "ibm00858" then 858
      else if encoding = "ibm860" then 860
      else if encoding = "ibm861" then 861
      else if encoding = "dos-862" then 862
      else if encoding = "ibm863" then 863
      else if encoding = "ibm864" then 864
      else if encoding = "ibm865" then 865
      else if encoding = "cp866" then 866
      else if encoding = "ibm869" then 869
      else if encoding = "ibm870" then 870
      else if encoding = "windows-874" then 874
      else if encoding = "cp875" then 875
      else if encoding = "shift_jis" then 932
      else if encoding = "gb2312" then 936
      else if encoding = "ks_c_5601-1987" then 949
      else if encoding = "big5" then 950
      else if encoding = "ibm1026" then 1026
      else if encoding = "ibm01047" then 1047
      else if encoding = "ibm01140" then 1140
      else if encoding = "ibm01141" then 1141
      else if encoding = "ibm01142" then 1142
      else if encoding = "ibm01143" then 1143
      else if encoding = "ibm01144" then 1144
      else if encoding = "ibm01145" then 1145
      else if encoding = "ibm01146" then 1146
      else if encoding = "ibm01147" then 1147
      else if encoding = "ibm01148" then 1148
      else if encoding = "ibm01149" then 1149
      else if encoding = "windows-1250" then 1250
      else if encoding = "windows-1251" then 1251
      else if encoding = "windows-1252" then 1252
      else if encoding = "windows-1253" then 1253
      else if encoding = "windows-1254" then 1254
      else if encoding = "windows-1255" then 1255
      else if encoding = "windows-1256" then 1256
      else if encoding = "windows-1257" then 1257
      else if encoding = "windows-1258" then 1258
      else if encoding = "johab" then 1361
      else if encoding = "macintosh" then 10000
      else if encoding = "x-mac-japanese" then 10001
      else if encoding = "x-mac-chinesetrad" then 10002
      else if encoding = "x-mac-korean" then 10003
      else if encoding = "x-mac-arabic" then 10004
      else if encoding = "x-mac-hebrew" then 10005
      else if encoding = "x-mac-greek" then 10006
      else if encoding = "x-mac-cyrillic" then 10007
      else if encoding = "x-mac-chinesesimp" then 10008
      else if encoding = "x-mac-romanian" then 10010
      else if encoding = "x-mac-ukrainian" then 10017
      else if encoding = "x-mac-thai" then 10021
      else if encoding = "x-mac-ce" then 10029
      else if encoding = "x-mac-icelandic" then 10079
      else if encoding = "x-mac-turkish" then 10081
      else if encoding = "x-mac-croatian" then 10082
      else if encoding = "x-chinese-cns" then 20000
      else if encoding = "x-cp20001" then 20001
      else if encoding = "x-chinese-eten" then 20002
      else if encoding = "x-cp20003" then 20003
      else if encoding = "x-cp20004" then 20004
      else if encoding = "x-cp20005" then 20005
      else if encoding = "x-ia5" then 20105
      else if encoding = "x-ia5-german" then 20106
      else if encoding = "x-ia5-swedish" then 20107
      else if encoding = "x-ia5-norwegian" then 20108
      else if encoding = "x-cp20261" then 20261
      else if encoding = "x-cp20269" then 20269
      else if encoding = "ibm273" then 20273
      else if encoding = "ibm277" then 20277
      else if encoding = "ibm278" then 20278
      else if encoding = "ibm280" then 20280
      else if encoding = "ibm284" then 20284
      else if encoding = "ibm285" then 20285
      else if encoding = "ibm290" then 20290
      else if encoding = "ibm297" then 20297
      else if encoding = "ibm420" then 20420
      else if encoding = "ibm423" then 20423
      else if encoding = "ibm424" then 20424
      else if encoding = "x-ebcdic-koreanextended" then 20833
      else if encoding = "ibm-thai" then 20838
      else if encoding = "koi8-r" then 20866
      else if encoding = "ibm871" then 20871
      else if encoding = "ibm880" then 20880
      else if encoding = "ibm905" then 20905
      else if encoding = "ibm00924" then 20924
      else if encoding = "euc-jp" then 20932
      else if encoding = "x-cp20936" then 20936
      else if encoding = "x-cp20949" then 20949
      else if encoding = "cp1025" then 21025
      else if encoding = "koi8-u" then 21866
      else if encoding = "iso-8859-2" then 28592
      else if encoding = "iso-8859-3" then 28593
      else if encoding = "iso-8859-4" then 28594
      else if encoding = "iso-8859-5" then 28595
      else if encoding = "iso-8859-6" then 28596
      else if encoding = "iso-8859-7" then 28597
      else if encoding = "iso-8859-8" then 28598
      else if encoding = "iso-8859-9" then 28599
      else if encoding = "iso-8859-13" then 28603
      else if encoding = "iso-8859-15" then 28605
      else if encoding = "x-europa" then 29001
      else if encoding = "iso-8859-8-i" then 38598
      else if encoding = "iso-2022-jp" then 50220
      else if encoding = "csiso2022jp" then 50221
      else if encoding = "iso-2022-jp" then 50222
      else if encoding = "iso-2022-kr" then 50225
      else if encoding = "x-cp50227" then 50227
      else if encoding = "euc-jp" then 51932
      else if encoding = "euc-cn" then 51936
      else if encoding = "euc-kr" then 51949
      else if encoding = "hz-gb-2312" then 52936
      else if encoding = "gb18030" then 54936
      else if encoding = "x-iscii-de" then 57002
      else if encoding = "x-iscii-be" then 57003
      else if encoding = "x-iscii-ta" then 57004
      else if encoding = "x-iscii-te" then 57005
      else if encoding = "x-iscii-as" then 57006
      else if encoding = "x-iscii-or" then 57007
      else if encoding = "x-iscii-ka" then 57008
      else if encoding = "x-iscii-ma" then 57009
      else if encoding = "x-iscii-gu" then 57010
      else if encoding = "x-iscii-pa" then 57011
      else TextEncoding.Utf8;

GetPowerQueryType = (dataType as text) =>
   if dataType = "boolean" then Logical.Type
   else if dataType = "number" then Number.Type
   else if dataType = "datetime" then DateTimeZone.Type
   else if dataType = "factor" then Text.Type
   else if dataType = "duration" then Duration.Type
   else if dataType = "text" then Text.Type
   else Text.Type ;

RunQuery = (query as text) =>
   let 
      url = base_url & "/query",
      body = "{ ""Sql"": """ & query & """ }",       
      Source = Json.Document(Web.Contents(url, [Headers = [#"Content-Type"="application/json"] , Content = Text.ToBinary(body) ]))
   in
      Source;

GetQueryStatus = (queryId as text) =>
   let 
      url = base_url & "/query/" & queryId & "/status",
      Source = Json.Document(Web.Contents(url, [IsRetry = true]))
   in
      if Source[Data][Status] = "Finished" then Source
      else if Source[Data][Status] = "Running" then Function.InvokeAfter( () => @GetQueryStatus(queryId), #duration(0,0,0,1))
      else if Source[Data][Status] = "Queued" then Function.InvokeAfter( () => @GetQueryStatus(queryId), #duration(0,0,0,5))
      else error Error.Record("Query Status Error","Running query is failed.","Query Status is " & Source[Data][Status]);

GetQuerySchema =  (queryId as text) =>
   let
      url = base_url & "/query/" & queryId & "/results/metadata/schema",
      Source =  Web.Contents(url, [IsRetry = true])
   in
      Source;
GetQueryResults = (queryId as text) =>
   let
      url = base_url & "/query/" & queryId & "/results/rawdata",
      Source = Binary.Decompress(
         Web.Contents(url, [IsRetry = true]),
         Compression.GZip
      )
   in
      Source;

// Data Source Kind description
Delika = [
   Authentication = [
      OAuth = [
         StartLogin = StartLogin,
         FinishLogin = FinishLogin
      ]
   ],
   Label = Extension.LoadString("DataSourceLabel")
];

Delika.SQL = [
   Authentication = [
      OAuth = [
         StartLogin = StartLogin,
         FinishLogin = FinishLogin
      ]
   ],
   Label = Extension.LoadString("DataSourceLabelSql")
];

Delika.QueryResults = [
   Authentication = [
      OAuth = [
         StartLogin = StartLogin,
         FinishLogin = FinishLogin
      ]
   ],
   Label = Extension.LoadString("DataSourceLabelQueryResults")
];

StartLogin = (resourceUrl, state, display) =>
   let
      AuthorizeUrl = Extension.LoadString("OAuthAuthorizeUri") & "?" & Uri.BuildQueryString([
         client_id = client_id,
         scope = scope,
         state = state,
         p = policy,
         response_type = "code",
         prompt = "login",
         redirect_uri = redirect_uri])
   in
      [
         LoginUri = AuthorizeUrl,
         CallbackUri = redirect_uri,
         WindowHeight = windowHeight,
         WindowWidth = windowWidth,
         Context = null
      ];

FinishLogin = (context, callbackUri, state) =>
   let
      Parts = Uri.Parts(callbackUri)[Query]
   in
      TokenMethod(Parts[code]);

TokenMethod = (code) =>
   let
      Response = Web.Contents(Extension.LoadString("OAuthTokenUri") & "?p=" & policy, [
         Content = Text.ToBinary(Uri.BuildQueryString([
            grant_type = "authorization_code",
            client_id = client_id,
            client_secret = client_secret,
            scope = scope,
            code = code,
            redirect_uri = redirect_uri])),
         Headers=[#"Content-type" = "application/x-www-form-urlencoded",#"Accept" = "application/json"]]),
      Parts = Json.Document(Response)
   in
      Parts;

// Data Source UI publishing description
Delika.Publish = [
   Beta = true,
   Category = "delika",
   ButtonText = { Extension.LoadString("ButtonTitle"), Extension.LoadString("ButtonHelp") },
   LearnMoreUrl = Extension.LoadString("LearnMoreUrl"),
   SourceImage = Delika.Icons,
   SourceTypeImage = Delika.Icons
];

Delika.SQL.Publish = [
   Beta = true,
   Category = "delika",
   ButtonText = { Extension.LoadString("ButtonTitleSql"), Extension.LoadString("ButtonHelpSql") },
   LearnMoreUrl = Extension.LoadString("LearnMoreUrl"),
   SourceImage = Delika.Icons,
   SourceTypeImage = Delika.Icons
];

Delika.QueryResults.Publish = [
   Beta = true,
   Category = "delika",
   ButtonText = { Extension.LoadString("ButtonTitleQueryResults"), Extension.LoadString("ButtonHelpQueryResults") },
   LearnMoreUrl = Extension.LoadString("LearnMoreUrl"),
   SourceImage = Delika.Icons,
   SourceTypeImage = Delika.Icons
];


Delika.Icons = [
   Icon16 = { Extension.Contents("Delika16.png"), Extension.Contents("Delika20.png"), Extension.Contents("Delika24.png"), Extension.Contents("Delika32.png") },
   Icon32 = { Extension.Contents("Delika32.png"), Extension.Contents("Delika40.png"), Extension.Contents("Delika48.png"), Extension.Contents("Delika64.png") }
];
